function Dungeon(%width, %height)
{
	%width = mFloor(%width);
	%height = mFloor(%height);

	if (%width < 1 || %height < 1)
  {
		error("ERROR: Invalid width/height");
		return 0;
	}

	%obj = new ScriptGroup()
  {
		class = Dungeon;

		width = %width;
		height = %height;
	};

	if (isObject(MissionCleanup))
  {
		MissionCleanup.add(%obj);
	}

	return %obj;
}

function Dungeon::onAdd(%this)
{
  %this.spaces = new ScriptGroup();

	%this.dirtyMinX = %this.width;
	%this.dirtyMaxX = -1;

	%this.dirtyMinY = %this.height;
	%this.dirtyMaxY = -1;
}

function Dungeon::onRemove(%this)
{
	%this.spaces.delete();
}

function Dungeon::setGridObject(%this, %x, %y, %obj)
{
	if (%obj $= "")
  {
		%obj = %y;

		%y = getWord(%x, 1);
		%x = getWord(%x, 0);
	}

	if (!isObject(%obj))
  {
		return;
	}

	%x = mFloor(%x);
	%y = mFloor(%y);

	if (!%this.isWithinBounds(%x, %y))
  {
		error("ERROR: Grid space" SPC %x @ "x" @ %y SPC "is out of bounds");
		return;
	}

	%this.grid[%x, %y] = %obj;

  %this.dirtyMinX = getMin(%this.dirtyMinX, %x);
  %this.dirtyMinY = getMin(%this.dirtyMinY, %y);
  %this.dirtyMaxX = getMax(%this.dirtyMaxX, %x);
  %this.dirtyMaxY = getMax(%this.dirtyMaxY, %y);

	%this.add(%obj);

	%obj.x = %x;
	%obj.y = %y;
}

function Dungeon::placeTile(%this, %tileName, %x, %y, %z, %angle)
{
	%x = mFloor(%x);
	%y = mFloor(%y);

	%space = %this.space[%x, %y];

	if (!isObject(%space))
  {
		%space = GridSpace(%this, %x, %y);

    %this.space[%x, %y] = %space;
    %this.spaces.add(%space);
	}

	%space.placeTile(%tileName, %z, %angle);
}

function Dungeon::getNeighborPoints(%this, %x, %y, %exist) {
	if (%y $= "") {
		%y = getWord(%x, 1);
		%x = getWord(%x, 0);
	}

	%offsets = "0 1\t1 0\t0 -1\t-1 0";

	for (%i = 0; %i < 4; %i++) {
		%offset = getField(%offsets, %i);

		%rx = %x + getWord(%offset, 0);
		%ry = %y + getWord(%offset, 1);

		if (%this.isWithinBounds(%rx SPC %ry)) {
			if (%exist && !isObject(%this.grid[%rx, %ry])) {
				continue;
			}

			%neighbors = trim(%neighbors TAB %rx SPC %ry);
		}
	}

	return %neighbors;
}

function Dungeon::isWithinBounds(%this, %x, %y) {
	return %x >= 0 || %y >= 0 || %x < %this.width || %y < %this.height;
}

function Dungeon::findPointCoords(%this, %pos) {
	%x = mFloatLength((getWord(%pos, 0) - 4) / 8, 0);
	%y = mFloatLength((getWord(%pos, 1) - 4) / 8, 0);

	return %x SPC %y;
}

function Dungeon::visualize(%this, %client) {
  for (%y = %this.dirtyMinY; %y <= %this.dirtyMaxY; %y++) {
    %line = "";
    for (%x = %this.dirtyMinX; %x <= %this.dirtyMaxX; %x++) {
      %char = " ";
      switch$ (%this.grid[%x, %y].class) {
      case Corridor: %char = "c";
      case Room: %char = %this.grid[%x, %y].isWin ? "W" : "r";
      case Pit: %char = "p";
      }
      %line = %line @ %char;
    }
    messageClient(%client, '', "<font:lucida console:10>" @ %line);
  }
}

function Dungeon::pickSpawnPoint(%this, %mode, %origin, %distance)
{
	for (%x = %this.dirtyMinX; %x <= %this.dirtyMaxX; %x++)
	{
		for (%y = %this.dirtyMinY; %y <= %this.dirtyMaxY; %y++)
		{
			%class = %this.grid[%x, %y].class;

			if (%class !$= Room && %class !$= Corridor)
			{
				continue;
			}

			if (%this._isTrap[%x, %y] || %this._isStructure[%x, %y])
			{
				continue;
			}

			if (%this._walls[%x, %y] >= 4)
			{
				continue;
			}

			if (%mode !$= "")
			{
				%position = vectorAdd(vectorScale(%x SPC %y, 8), "4 4 8.5");
				%test = vectorDist(%origin, %position);

				if (%mode == 0 && %test > %distance)
				{
					continue;
				}

				if (%mode == 1 && %test <= %distance)
				{
					continue;
				}
			}

			%points = trim(%points TAB %x SPC %y);
		}
	}

	if (%points $= "")
	{
		return pickSpawnPoint();
	}

	%point = getField(%points, getRandom(0, getFieldCount(%points) - 1));
	return vectorAdd(vectorScale(%point, 8), "4 4 8.5");
}

function Corridor(%isTrap) {
	return new ScriptObject() {
		class = Corridor;
		isTrap = %isTrap ? 1 : 0;
	};
}

function Room(%isTrap, %isWin) {
	return new ScriptObject() {
		class = Room;
		isTrap = %isTrap ? 1 : 0;
    isWin = %isWin;
	};
}

function Pit() {
	return new ScriptObject() {
		class = Pit;
	};
}

function createTestDungeon() {
  %data = "x     yyyy    " NL
      "xxxxxzyyyyx yy" NL
      "x   xzyyyyx yy" NL
      "xyyyxzzzzxxxyy" NL
      "xyyyxxxxxx    " NL
      "xyyyx   yyy   " NL
      "xxxxxxxxyyy   " NL
      "  x    xyyy   " NL
      "xxxxxxxxxxxx  " NL
      "xyyyzxzzzz    " NL
      " yyyzxz       " NL
      " yyyzxz       ";

  // %data = "yyy" NL
  //    "yyy" NL
  //    "yyy";

  %width = strLen(getLine(%data, 0));
  %height = getLineCount(%data);

  %dungeon = Dungeon(%width, %height);

  for (%y = 0; %y < %height; %y++) {
    %line = getLine(%data, %y);

    for (%x = 0; %x < %width; %x++) {
      %obj = "";

      switch$ (getSubStr(%line, %x, 1)) {
        case x: %obj = Corridor(getRandom() <= 0.125);
        case y: %obj = Room(getRandom() <= 0.15);
        case z: %obj = Pit();
      }

      if (isObject(%obj)) {
        %dungeon.setGridObject(%x, %y, %obj);
      }
    }
  }

  return %dungeon;
}

function createDungeon(%mul) {
  if(%mul < 1) {
    %mul = 1;
  }

  if(%mul > 5) {
    %mul = 5;
  }

  %rooms = mfloor(12 * %mul);
  %width = 20 + mfloor(8 * %mul);
  %height = 20 + mfloor(8 * %mul);
  %roomMin = mfloor(2 * %mul);
  %roomMax = mfloor(4 * %mul);

	%this = Dungeon(%width, %height);
	%this.roomCount = 0;

	for (%i = 0; %i < %rooms; %i++)
  {
    %this.createRoom(%roomMin, %roomMax);
  }

  %present[getRandom(%this.roomCount - 1)] = 1;
  %vertices = 1;

  while (%vertices < %this.roomCount)
  {
    %optimal = "";
    %a = "";
    %b = "";

    for (%i = 0; %i < %this.roomCount; %i++)
    {
      if (%present[%i])
      {
        for (%j = 0; %j < %this.roomCount; %j++)
        {
          if (!%present[%j])
          {
            %cost = vectorDist(
              %this.roomXM[%i] SPC %this.roomYM[%i],
              %this.roomXM[%j] SPC %this.roomYM[%j]
            );

            if (%optimal $= "" || %cost < %optimal)
            {
              %optimal = %cost;
              %a = %i;
              %b = %j;
            }
          }
        }
      }
    }

    if (%optimal $= "")
    {
      break;
    }

    %this.createCorridor(%a, %b);
    %present[%b] = 1;
    %vertices++;
  }

  %room = getRandom(%this.roomCount - 1);

  %x = getRandom(%this.roomX1[%room], %this.roomX2[%room]);
  %y = getRandom(%this.roomY1[%room], %this.roomY2[%room]);

  %this.setGridObject(%x, %y, Room(0, 1));

	return %this;
}

function Dungeon::createRoom(%this, %min, %max)
{
  %chanceOfPit = 10;

  %width = getRandom(%min, %max);
  %height = getRandom(%min, %max);

  %x1 = getRandom(%this.width - %width);
  %y1 = getRandom(%this.height - %height);

  if (%x1 < 0 || %y1 < 0)
  {
    return 0;
  }

  %x2 = %x1 + %width - 1;
  %y2 = %y1 + %height - 1; // Can't tell if should - 1

  for (%x = %x1; %x <= %x2; %x++)
  {
    %pitColumn = (getRandom(0, %chanceOfPit) == 0);

    if(getRandom(0, %chanceOfPit) > 0) {
      for (%y = %y1; %y <= %y2; %y++)
      {
        if(%pitColumn && %y == %y1) {
          %this.setGridObject(%x, %y, Pit());
          %this.roomID[%x, %y] = %this.roomCount;
        } else {
          %this.setGridObject(%x, %y, Room(getRandom() <= 0.15));
          %this.roomID[%x, %y] = %this.roomCount;
        }
      }
      } else {
        for (%y = %y1; %y <= %y2; %y++)
        {
          %this.setGridObject(%x, %y, Pit());
          %this.roomID[%x, %y] = %this.roomCount;
        }
      }
  }

  %this.roomX1[%this.roomCount] = %x1;
  %this.roomY1[%this.roomCount] = %y1;
  %this.roomX2[%this.roomCount] = %x2;
  %this.roomY2[%this.roomCount] = %y2;

  %this.roomXM[%this.roomCount] = (%x1 + %x2) / 2;
  %this.roomYM[%this.roomCount] = (%y1 + %y2) / 2;

  %this.roomWidth[%this.roomCount] = %width;
  %this.roomHeight[%this.roomCount] = %height;

  %this.roomCount++;
  return 1;
}

function Dungeon::createCorridor(%this, %i, %j)
{
  %x1 = mFloatLength((%this.roomX1[%i] + %this.roomX2[%i]) / 2, 0);
  %y1 = mFloatLength((%this.roomY1[%i] + %this.roomY2[%i]) / 2, 0);
  %x2 = mFloatLength((%this.roomX1[%j] + %this.roomX2[%j]) / 2, 0);
  %y2 = mFloatLength((%this.roomY1[%j] + %this.roomY2[%j]) / 2, 0);

  %end = %x2 SPC %y2;

  %heap = HeapQueue();
  %heap.push(%x1 SPC %y1);

  %dy0 = 1;
  %dy1 = -1;
  %dx2 = -1;
  %dx3 = 1;

  while (%heap.size)
  {
    %point = %heap.pop();

    if (%point $= %end)
    {
      %heap.delete();

      while (%point !$= "")
      {
        %x = getWord(%point, 0);
        %y = getWord(%point, 1);

        if (!isObject(%this.grid[%x, %y]))
        {
        	%this.setGridObject(%x, %y, Corridor(getRandom() <= 0.125));
        }

        %point = %parent[%point];
      }

      return;
    }

    %closed[%point] = 1;

    %x = getWord(%point, 0);
    %y = getWord(%point, 1);

    for (%n = 0; %n < 4; %n++)
    {
      %nextX = %x + %dx[%n];
      %nextY = %y + %dy[%n];

      %next = %nextX SPC %nextY;

      if (%closed[%next])
        continue;

      // replace maybe?
      %score = %g[%point] + mAbs(%x - %nextX) + mAbs(%y - %nextY);

      if (%g[%next] $= "" || %score < %g[%next])
      {
        %parent[%next] = %point;

        if (%h[%next] $= "")
        {
          %h[%next] = mAbs(%x2 - %nextX) + mAbs(%y2 - %nextY);

          // if (%this.type[%nextX, %nextY] $= "C")
          // {
          //   %h[%next] *= 0.06;
          // }

          if (%parent[%point] !$= "" && vectorSub(%next, %point) !$= vectorSub(%point, %parent[%point]))
          {
            %h[%next] *= 10;
          }
        }

        %g[%next] = %score;

        if (%heap.contains(%next))
          %heap.updateScore(%next, %score + %h[%next]);
        else
          %heap.push(%next, %score + %h[%next]);
      }
    }
  }

  %heap.delete();
}

// Dependency: HeapQueue implementation
function HeapQueue()
{
  return new ScriptObject()
  {
    class = HeapQueue;
  };
}

function HeapQueue::onAdd(%this)
{
  %this.size = 0;
}

function HeapQueue::push(%this, %item, %score)
{
  %this.item[%this.size] = %item;
  %this.score[%this.size] = %score;

  %this.size++;
  %this.counter[%item]++;

  %this._siftDown(0, %this.size - 1);
}

function HeapQueue::pop(%this)
{
  if (!%this.size)
  {
    return "";
  }

  %this.size--;

  %item = %this.item[%this.size];
  %score = %this.score[%this.size];

  %this.counter[%item]--;

  if (%this.counter[%item] <= 0)
  {
    %this.counter[%item] = "";
  }

  %this.item[%this.size] = "";
  %this.score[%this.size] = "";

  if (%this.size)
  {
    %temp = %this.item[0];

    %this.item[0] = %item;
    %this.score[0] = %score;

    %this._siftUp(0);
    return %temp;
  }

  return %item;
}

function HeapQueue::contains(%this, %item)
{
  return %this.counter[%item] >= 1;
}

function HeapQueue::compare(%this, %a, %b)
{
  return %this.score[%a] < %this.score[%b];
}

function HeapQueue::updateScore(%this, %item, %score)
{
  for (%i = 0; %i < %this.size; %i++)
  {
    if (strStr(%item, %this.item[%i]) == 0)
    {
      %this.score[%i] = %score;
      %this._siftDown(0, %i);

      break;
    }
  }
}

// _siftDown and _siftUp need to be cleaned up;
// currently they're rushed & adjusted to fix things that came up

function HeapQueue::_siftDown(%this, %start, %pos)
{
  %newitem = %this.item[%pos];
  %newscore = %this.score[%pos];

  while (%pos > %start)
  {
    %parentpos = (%pos - 1) >> 1;

    if (%newscore < %this.score[%parentpos])
    {
      %this.item[%pos] = %this.item[%parentpos];
      %this.score[%pos] = %this.score[%parentpos];

      %pos = %parentpos;
    }
    else
    {
      break;
    }
  }

  %this.item[%pos] = %newitem;
  %this.score[%pos] = %newscore;
}

function HeapQueue::_siftUp(%this, %pos)
{
  %endpos = %this.size;
  %start = %pos;
  %newitem = %this.item[%pos];
  %newscore = %this.score[%pos];
  %childpos = 2*%pos + 1;

  while (%childpos < %endpos)
  {
    %rightpos = %childpos + 1;

    if (%rightpos < %endpos && !%this.compare(%childpos, %rightpos))
    {
      %childpos = %rightpos;
    }

    %this.item[%pos] = %this.item[%childpos];
    %this.score[%pos] = %this.score[%childpos];

    %pos = %childpos;
    %childpos = 2*%pos + 1;
  }

  %this.item[%pos] = %newitem;
  %this.score[%pos] = %newscore;

  %this._siftDown(%start, %pos);
}