datablock AudioProfile(GameOver)
{
  fileName = "Add-Ons/GameMode_King_Of_Ghosts/res/GameOver.wav";
  description = AudioDefault3D;
  preload = 1;
};

datablock AudioProfile(BeginLevel1)
{
  fileName = "Add-Ons/GameMode_King_Of_Ghosts/res/HordeIncoming1.wav";
  description = AudioDefault3D;
  preload = 1;
};

datablock AudioProfile(BeginLevel2)
{
  fileName = "Add-Ons/GameMode_King_Of_Ghosts/res/HordeIncoming2.wav";
  description = AudioDefault3D;
  preload = 1;
};

package KOGPackage
{
  function MiniGameSO::onRemove(%this)
  {
    if (isObject(%this.kogSession))
    {
      %this.kogSession.delete();
    }

    Parent::onRemove(%this);
  }

  function MiniGameSO::reset(%this, %client, %finished)
  {
    for ( %i = 0 ; %i < %this.numMembers ; %i++ )
    {
      %this.member[ %i ].isDead = "";
    }

    if (%this.owner != 0)
    {
      return Parent::reset(%this, %client);
    }

    if (!%finished && isObject(%this.kogSession))
    {
      %this.kogSession.delete();
    }

    if (%finished)
    {
      %this.lastResetTime -= 5000;
    }

    Parent::reset(%this, %client);

    if (%this.numMembers && !isObject(%this.kogSession))
    {
      %this.kogSession = KOGSession(%this);
    }
  }

  function MiniGameSO::scheduleReset(%this, %time) {
    serverPlay2D(GameOver);

    %baseGOText = "<font:palatino linotype:48><color:FF0000>GAME OVER\n";
    %baseGOText = %baseGOText @ "<font:palatino linotype:32><color:FFFFFF>Perished on floor " @ %this.kogSession.level @ "\n";
    %baseGOText = %baseGOText @ "<font:palatino linotype:24>\n<color:FFFFFF>HINT:\n";

    %selectHint = getRandom(0, 6);
    %selectedHint = "lolwtf error";

    switch(%selectHint) {
      case 0:
        %selectedHint = "Look at the screen while playing";
      case 1:
        %selectedHint = "This gamemode is not suitable for people over the age 115";
      case 2:
        %selectedHint = "The ghosts ALL have superior weapons to you. If you can, try to avoid them";
      case 3:
        %selectedHint = "Stop dying";
      case 4:
        %selectedHint = "If you're seeing this, you're playing it wrong";
      case 5:
        %selectedHint = "Use found teleporters to your advantage";
      case 6:
        %selectedHint = "Tripwires are the barely visible bricks on the ground - keep an eye out for them";
    }

    %this.centerPrintAll(%baseGOText @ %selectedHint, 0);

    return Parent::scheduleReset(%this, %time);
  }

  function MiniGameSO::addMember(%this, %client)
  {
    Parent::addMember(%this, %client);

    if (%this.owner == 0 && %this.numMembers && !isObject(%this.kogSession))
    {
      %this.reset(0);
    }
  }

  function MiniGameSO::removeMember(%this, %client)
  {
    Parent::removeMember(%this, %client);

    %client.isDead = "";

    if (!%this.numMembers && isObject(%this.kogSession))
    {
      %this.kogSession.delete();
    }
  }

  function MiniGameSO::pickSpawnPoint(%this)
  {
    if (isObject(%this.kogSession))
    {
      %area = %this.kogSession.pickSpawnPoint(0, %this.kogSession.dungeon.spawnArea, 40);
      %area = vectorAdd(%area, getRandom() * 2 - 1 SPC getRandom() * 2 - 1);

      return %area;
    }
    else
    {
      return Parent::pickSpawnPoint(%this);
    }
  }

  function GameConnection::spawnPlayer(%this)
  {
    if (!%this.miniGame.kogSession.dungeon._isBuilding && !%this.isDead)
    {
      Parent::spawnPlayer(%this);

      // give random loadout
      %player = %this.player;
      %numItems = getRandom(1, 5);

      for(%i = 0; %i < %numItems; %i++) {
        %player.addRandomItemBlacklisted(%this);
      }
    }
  }

  function GameConnection::instantRespawn(%this)
  {
    if (!isObject(%this.miniGame.kogSession) || !isObject(%this.player))
    {
      return Parent::instantRespawn(%this);
    }

    %area = %this.miniGame.kogSession.pickSpawnPoint();
    %area = vectorAdd(%area, getRandom() * 2 - 1 SPC getRandom() * 2 - 1);

    %this.player.setTransform(%area);
    %this.player.setWhiteOut(0.5);

    %this.player.spawnExplosion(SpawnExplosion, 1);
  }

  function miniGameCanDamage(%obj, %col)
  {
    %mini1 = getMiniGameFromObject(%obj);
    %mini2 = getMiniGameFromObject(%col);

    if (%mini1 != $DefaultMiniGame || %mini2 != $DefaultMiniGame)
    {
      return Parent::miniGameCanDamage(%obj, %col);
    }

    if (!(%col.getType() & $TypeMasks::PlayerObjectType))
    {
      return Parent::miniGameCanDamage(%obj, %col);
    }

    if (%col.client == %obj)
    {
      return 1;
    }

    if (%obj.getType() & $TypeMasks::PlayerObjectType)
    {
      %ghost1 = %obj.isGhost;
    }
    else
    {
      %className = %obj.getClassName();

      switch$ (%className)
      {
        case "GameConnection": %ghost1 = %obj.player.isGhost;
      }
    }

    %ghost2 = %col.isGhost ? 1 : 0;

    if ((%ghost1 && %ghost2) || (!%ghost1 && !%ghost2))
    {
      return 0;
    }

    return 1;
  }

  function gameConnection::createPlayer( %this, %transform )
  {
    %this.isDead = "";
    return parent::createPlayer( %this, %transform );
  }

  function gameConnection::onDeath( %this, %pl, %cl, %type, %loc )
  {
    parent::onDeath( %this, %pl, %cl, %type, %loc );
    messageClient( %this, 'MsgYourSpawn' );

    %this.isDead = true;
    %this.centerPrint( "\c5You are dead.", 2 );

    %this.miniGame.kogSession.survivorCheck();
  }

  function Observer::onTrigger(%this,%obj,%trigger,%state)
  {
    %mini = $DefaultMiniGame;

    if(%this.currentObserved $= "") {
      %this.currentObserved = -1;
    }

    %this.currentObserved++;

    %loops = 0;
    while(!isObject(%mini.member[%this.currentObserved].player)) {
      %this.currentObserved++;
      %loops += 1;

      if(%this.currentObserved >= %mini.numMembers) {
        %this.currentObserved = 0;
      }

      if(%loops >= %mini.numMembers) {
        return;
      }
    }

    %this.setMode("corpse", %mini.member[%this.currentObserved].player);
    %client.setControlObject(%this);

    %obj.bottomPrint("Spectating " + %mini.member[%this.currentObserved].name, 1, true);

    Parent::onTrigger(%this, %obj, %trigger, %state);
  }
};

activatePackage("KOGPackage");